#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
isoenum.openbabel
~~~~~~~~~~~~~~~~~

This module provides :func:`~isoenum.openbabel.convert` to convert
between different file formats used in molecular modeling and computational
chemistry (e.g. ``InChI``, ``SMILES``, ``Molfile``, etc.).
"""

import re
import subprocess

from packaging import version

try:
    from subprocess import DEVNULL
except ImportError:
    import os

    DEVNULL = open(os.devnull, "wb")


def _test_openbabel(required_openbabel_version="2.4.1"):
    """Test if Open Babel software is installed.
    
    :return: None.
    :rtype: :py:obj:`None`
    """
    try:
        openbabel_version_test = subprocess.check_output(["obabel", "-V"])
        openbabel_version = openbabel_version_test.decode("utf-8")

        version_pattern = re.compile("\d{1,2}\.\d{1,2}\.\d{1,2}")
        openbabel_version_number = re.search(version_pattern, openbabel_version)

        if not openbabel_version_number:
            raise SystemExit(
                "Open Babel version information cannot be found: {}".format(
                    openbabel_version
                )
            )
        else:
            if version.parse(openbabel_version_number.group()) < version.parse(
                required_openbabel_version
            ):
                raise SystemExit(
                    "Open Babel version {} is required.".format(
                        required_openbabel_version
                    )
                )

    except OSError:
        raise SystemExit(
            "Open Babel software is not installed, exiting. "
            "See installation instructions to get Open Babel "
            "software for your operating system:"
            "http://openbabel.org/wiki/Get_Open_Babel"
        )


def convert(input_file_path, output_file_path, input_format, output_format, **options):
    """Convert between formats using Open Babel.
    
    :param str input_file_path: Path to input file.
    :param str output_file_path: Path to output file.
    :param str input_format: Input file format.
    :param str output_format: Output file format.
    :param options: Additional key-value options to pass to Open Babel.
    :return: None.
    :rtype: :py:obj:`None` 
    """
    _test_openbabel()

    cmd = [
        "obabel",
        "-i{}".format(input_format),
        "{}".format(input_file_path),
        "-o{}".format(output_format),
        "-O{}".format(output_file_path),
    ]

    if options:
        for option in options.values():
            cmd.extend(option.split())

    subprocess.call(cmd, shell=False, stdout=DEVNULL, stderr=subprocess.STDOUT)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""
isoenum.labeling
~~~~~~~~~~~~~~~~

This module contains a function to generate a labeling schema 
based on provided cli parameters.
"""

import itertools
from collections import defaultdict
from collections import Counter


def create_labeling_schema(complete_labeling_schema, ignore_existing_isotopes,
                           all_iso, specific_iso, existing_iso, enumerate_iso,
                           isotopes_conf, ctfile):
    """Create labeling schema.

    :param bool complete_labeling_schema: Specifies if default isotopes to be added to isotopic layer.
    :param bool ignore_existing_isotopes: Specifies if will ignore existing isotopic layer.
    :param dict all_iso: Atom specific isotopes `--all` option.
    :param dict specific_iso: Atom number specific isotopes from `--specific` option.
    :param dict existing_iso: Atom number specific isotopes from ``Molfile``.
    :param list enumerate_iso: List of isotopes from `--enumerate` option. 
    :param dict isotopes_conf: Default isotopes.
    :param ctfile: Instance of ``Molfile``. 
    :type ctfile: :class:`~ctfile.ctfile.Molfile`
    :return: Labeling schema.
    :rtype: :py:class:`list`
    """
    allowed_atom_symbols = [atom.atom_symbol for atom in ctfile.atoms]
    positions = [atom.atom_number for atom in ctfile.atoms]
    ctfile_position_atom = dict(zip(positions, allowed_atom_symbols))
    starting_iso = {}

    # first, use all atom specification to create labeling schema
    if all_iso:
        starting_iso.update(all_iso)

    # second, use specific atom bspecification to create labeling schema
    if specific_iso:
        starting_iso.update(specific_iso)

    # third, keep the isotopic layer from original CTfile
    if not ignore_existing_isotopes:
        if existing_iso:
            starting_iso.update(existing_iso)

    if not enumerate_iso:
        # add default isotopes if need to create full labeling schema
        if complete_labeling_schema:
            default_iso = _default_isotopes(ctfile=ctfile, isotopes_conf=isotopes_conf, current_iso=starting_iso)
            starting_iso.update(default_iso)

        labeling_schema = sorted(starting_iso.values(), key=lambda k: int(k['atom_number']))
        yield labeling_schema

    else:
        isotopes_per_atom = defaultdict(set)

        for entry in enumerate_iso:
            isotopes_per_atom[entry['atom_symbol']].add(entry['isotope'])
            isotopes_per_atom[entry['atom_symbol']].add(None)  # None designates that isotope is not specified

        list_of_isotopes_per_position = defaultdict(list)
        for position, atom in ctfile_position_atom.items():
            list_of_isotopes_per_position[position].extend(isotopes_per_atom.get(atom, [None]))

        for entry in starting_iso.values():
            list_of_isotopes_per_position[entry['atom_number']] = [entry['isotope']]

        list_of_isotopes_per_position_sorted_by_position = [list_of_isotopes_per_position[iso] for iso in
                                                            sorted(list_of_isotopes_per_position, key=int)]

        labeling_product = itertools.product(*list_of_isotopes_per_position_sorted_by_position)
        for prod in labeling_product:
            labeling_schema = {}

            isotopes_per_atom = ['{}-{}'.format(atom, isotope) for atom, isotope in zip(allowed_atom_symbols, prod)]
            counter = Counter(isotopes_per_atom)

            valid_labeling = [
                True
                if entry['min'] <= counter['{}-{}'.format(entry['atom_symbol'], entry['isotope'])] <= entry['max']
                else False
                for entry in enumerate_iso
            ]

            if all(valid_labeling):
                for atom, isotope, position in zip(allowed_atom_symbols, prod, positions):
                    if isotope:
                        labeling_schema[position] = {'atom_symbol': atom, 'isotope': isotope, 'atom_number': position}

            if complete_labeling_schema:
                default_iso = _default_isotopes(ctfile=ctfile, isotopes_conf=isotopes_conf, current_iso=labeling_schema)
                labeling_schema.update(default_iso)

            labeling_schema = sorted(labeling_schema.values(), key=lambda k: int(k['atom_number']))
            if labeling_schema:
                yield labeling_schema


def _default_isotopes(ctfile, isotopes_conf, current_iso):
    """Create dictionary with default isotopes.

    :param ctfile: Instance of ``Molfile``. 
    :type ctfile: :class:`~ctfile.ctfile.Molfile`.
    :param dict isotopes_conf: Default isotopes.
    :param dict current_iso: Current isotopes to which default isotopes will be added. 
    :return: Default isotopes.
    :rtype: :py:class:`dict`
    """
    default_iso = {}
    allowed_atom_symbols = [atom.atom_symbol for atom in ctfile.atoms]
    positions = [atom.atom_number for atom in ctfile.atoms]
    ctfile_position_atom = dict(zip(positions, allowed_atom_symbols))

    for atom_number in positions:
        if atom_number not in current_iso:
            atom = ctfile_position_atom[atom_number]
            isotope = isotopes_conf[atom]['default']
            default_iso[atom_number] = {'atom_symbol': atom, 'isotope': isotope, 'atom_number': atom_number}
    return default_iso



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""

POST /InChI.asmx/InChIToMol HTTP/1.1
Host: www.chemspider.com
Content-Type: application/x-www-form-urlencoded
Content-Length: length

inchi=string

"""

import os
import sys
import json
import xml.etree.ElementTree as ET

import requests
import pdir

with open("{}/conf/chemspider.json".format(os.path.abspath(os.path.dirname(__file__))), "r") as infile:
    chemspider_conf = json.load(infile)


def inchi_to_mol(inchi, chemspider_conf):
    """Convert from inchi string or file to molfile string or file. 
    
    :param inchi: InChI string or path to file containing InChI string.
    :param str url: URL path to  
    :param dict chemspider_conf:
    :return: 
    :rtype: str
    """
    if os.path.exists(inchi):
        with open(inchi, 'r') as inf:
            inchi = inf.read().strip()

    if not inchi.startswith('InChI='):
        inchi = 'InChI={}'.format(inchi)

    url = chemspider_conf['inchi_to_mol']['url']
    headers = chemspider_conf['inchi_to_mol']['headers']
    payload = {'inchi': inchi}

    r = requests.post(url, data=payload, headers=headers)

    if r.ok:
        root = ET.fromstring(r.text)
        return root.text
    return ""


def mol_to_inchi(inchi):
    pass


if __name__ == '__main__':
    script = sys.argv.pop(0)
    inchi = sys.argv.pop(0)

    mol_str = inchi_to_mol(inchi, chemspider_conf)
    print(mol_str)


#!/usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
isoenum.conf
~~~~~~~~~~~~

This module provides processing of configuration files necessary for 
isotopic enumerator.
"""

import os
import json


this_directory = os.path.abspath(os.path.dirname(__file__))
isotopes_path = '{}/config_files/isotopes.json'.format(this_directory)
output_formats_path = '{}/config_files/formats.json'.format(this_directory)

with open(isotopes_path, 'r') as infile:
    isotopes_conf = json.load(infile)

with open(output_formats_path, 'r') as infile:
    output_formats_conf = set(json.load(infile))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
isoenum.nmr
~~~~~~~~~~~

This module provides descriptions of coupling combinations that
could be observed within NMR experiments.
"""

import itertools

import more_itertools

from . import utils


class ResonanceAtom(object):
    """Resonance atom - part of the coupling path."""

    def __init__(self, atom, isotope):
        """Resonance atom initializer.

        :param atom: Atom.
        :type atom: :class:`ctfile.Atom.`
        :param str isotope: Absolute mass of the atom.
        """
        self.atom = atom
        self.isotope = isotope
        self.atom_number = atom.atom_number
        self.atom_symbol = atom.atom_symbol

    def __str__(self):
        """String representation of resonance atom.
        
        :return: String representation of resonance atom.
        :rtype: :py:class:`str`
        """
        return '{}{}{}'.format(self.isotope, self.atom_symbol, self.atom_number)

    def __repr__(self):
        """String representation of resonance atom.

        :return: String representation of resonance atom.
        :rtype: :py:class:`str`
        """
        return str(self)


class Coupling(object):
    """Coupling provides information on the connectivity of molecules."""

    def __init__(self, coupling_path=None, nmr_active_atoms=None, subset_atoms=None):
        """Coupling initializer.
        
        :param list coupling_path: List of atoms that are responsible for observed coupling type.
        :param dict nmr_active_atoms: Atom-isotope pairs that describe labeling.
        :param dict subset_atoms: Atom-isotope pairs used for subset generation.
        """
        if coupling_path:
            self.coupling_path = [sorted(atom_group, key=lambda atom: int(atom.atom_number))
                                  for atom_group in coupling_path]
        else:
            self.coupling_path = coupling_path

        self.nmr_active_atoms = nmr_active_atoms
        self.subset_atoms = subset_atoms

    @property
    def coupling_type(self):
        """Coupling type.
        
        :return: Coupling type.
        :rtype: :py:class:`str`
        """
        return self.__class__.__qualname__

    @property
    def subset(self):
        """Generate coupling path subsets.
        
        :return: Coupling path subsets.
        :rtype: :py:class:`list`
        """
        subset_coupling_path = []

        for atoms_group in self.coupling_path:
            new_atoms_group = []
            for rsize in range(1, len(atoms_group)+1):
                subset = [self._atom_substitution(original_atoms=atoms_group, new_atoms=list(combination))
                          for combination in itertools.combinations(atoms_group, rsize)]
                new_atoms_group.extend(subset)
            subset_coupling_path.append(new_atoms_group)

        subset_couplings = [self.__class__(coupling_path=list(coupling_path),
                                           nmr_active_atoms=self.nmr_active_atoms,
                                           subset_atoms=self.subset_atoms)
                            for coupling_path in itertools.product(*subset_coupling_path)]
        return subset_couplings

    def _atom_substitution(self, original_atoms, new_atoms):
        """Helper function to add missing atom(s) to group of atoms during subset generation.
        
        :param list original_atoms: Group of atoms as part of coupling path before subset.
        :param list new_atoms: Group of atoms as part of coupling path after subset.
        :return: Group of atoms as part of coupling path after subset
        :rtype: :py:class:`list`
        """
        if len(new_atoms) == len(original_atoms):
            return original_atoms
        else:
            atoms_difference = set(new_atoms).symmetric_difference(original_atoms)
            for atom in atoms_difference:
                new_atoms.append(ResonanceAtom(atom=atom, isotope=self.subset_atoms.get(atom.atom_symbol, '')))
            return new_atoms

    def is_resonance_compatible(self, resonance):
        """Test if coupling is compatible with resonance.

        :param resonance: Subclass of :class:`~isoenum.nmr.Coupling`.
        :type resonance: :class:`~isoenum.nmr.Coupling`
        :return: True if compatible, False otherwise.
        :rtype: :py:obj:`True` or `False`
        """
        if set(resonance.hydrogen_coupling_path_repr[0]) == set(more_itertools.flatten(self.hydrogen_coupling_path_repr)):
            return True
        else:
            return False

    @classmethod
    def couplings(cls, atom):
        """Generate list of possible couplings for a given atom type.
        
        :param atom: Atom type.
        :type atom: :class:`ctfile.Atom`
        :return: List of couplings.
        :rtype: :py:class:`list`
        """
        return NotImplementedError('Subclass must implement method.')

    @property
    def name(self):
        """Specific coupling name that indicates interacting atoms.

        :return: Specific coupling name.
        :rtype: :py:class:`str`
        """
        return '[{}:{}]{}'.format(
            ','.join(['{}{}{}'.format(atom.isotope, atom.atom_symbol, atom.atom_number) for atom in self.coupling_path[0]]),
            ','.join(['{}{}{}'.format(atom.isotope, atom.atom_symbol, atom.atom_number) for atom in self.coupling_path[-1]]),
            self.coupling_type)

    @property
    def hydrogen_coupling_path(self, atom_symbol='H'):
        """Coupling path that consists of hydrogen atoms.

        :return: Coupling path.
        :rtype: :py:class:`list`
        """
        return list(filter(lambda lst: bool(lst),
                           [[atom for atom in atom_group if atom.atom_symbol == atom_symbol] for atom_group in
                            self.coupling_path]))

    @property
    def hydrogen_coupling_path_repr(self):
        """Hydrogen coupling path representation.

        :return: List of hydrogen groups representing hydrogen coupling path.
        :rtype: :py:class:`list`
        """
        return [['{}{}{}'.format(atom.isotope, atom.atom_symbol, atom.atom_number) for atom in atom_group]
                for atom_group in self.hydrogen_coupling_path]

    def __eq__(self, other):
        """Comparison of couplings based on atom and isotope specific coupling name."""
        return self.name == other.name

    def __ne__(self, other):
        """Comparison of couplings based on atom and isotope specific coupling name."""
        return self.name != other.name

    def __str__(self):
        """String representation of coupling.
        
        :return: String representation of coupling.
        :rtype: :py:class:`str`.
        """
        return '{}(coupling_path={})'.format(self.__class__.__qualname__, self.coupling_path)

    def __repr__(self):
        """String representation of coupling.
        
        :return: String representation of coupling.
        :rtype: :py:class:`str`.
        """
        return str(self)


class J1CH(Coupling):
    """J1CH coupling type: C-H."""

    def __init__(self, coupling_path=None, nmr_active_atoms=None, subset_atoms=None):
        super(J1CH, self).__init__(coupling_path=coupling_path,
                                   nmr_active_atoms=nmr_active_atoms,
                                   subset_atoms=subset_atoms)

    def couplings(self, carbon_atom):
        """Generate list of possible J1CH couplings.

        :param carbon_atom: Carbon atom.
        :type atom: :class:`ctfile.Atom`
        :return: List of couplings.
        :rtype: :py:class:`list`
        """
        couplings = []
        if len(carbon_atom.neighbor_hydrogen_atoms) > 0:
            coupling_path = [
                [ResonanceAtom(atom=atom, isotope=self.nmr_active_atoms.get(atom.atom_symbol, '')) for atom in carbon_atom.neighbor_hydrogen_atoms],
                [ResonanceAtom(atom=carbon_atom, isotope=self.nmr_active_atoms.get(carbon_atom.atom_symbol, ''))]
            ]
            couplings.append(self.__class__(coupling_path=coupling_path,
                                            nmr_active_atoms=self.nmr_active_atoms,
                                            subset_atoms=self.subset_atoms))
        return couplings


class J2HH(Coupling):
    """J2HH coupling type: H-C-H."""

    def __init__(self, coupling_path=None, nmr_active_atoms=None, subset_atoms=None):
        super(J2HH, self).__init__(coupling_path=coupling_path,
                                   nmr_active_atoms=nmr_active_atoms,
                                   subset_atoms=subset_atoms)

    def couplings(self, carbon_atom):
        """Generate list of possible J2HH couplings.

        :param carbon_atom: Carbon atom.
        :type atom: :class:`ctfile.Atom`
        :return: List of couplings.
        :rtype: :py:class:`list`
        """
        couplings = []
        if len(carbon_atom.neighbor_hydrogen_atoms) == 2:
            coupling_path = [
                [ResonanceAtom(atom=carbon_atom.neighbor_hydrogen_atoms[0], isotope=self.nmr_active_atoms.get(carbon_atom.neighbor_hydrogen_atoms[0].atom_symbol, ''))],
                [ResonanceAtom(atom=carbon_atom, isotope=self.nmr_active_atoms.get(carbon_atom.atom_symbol, ''))],
                [ResonanceAtom(atom=carbon_atom.neighbor_hydrogen_atoms[1], isotope=self.nmr_active_atoms.get(carbon_atom.neighbor_hydrogen_atoms[1].atom_symbol, ''))]
            ]
            couplings.append(self.__class__(coupling_path=coupling_path,
                                            nmr_active_atoms=self.nmr_active_atoms,
                                            subset_atoms=self.subset_atoms))
        return couplings


class J3HH(Coupling):
    """J3HH coupling type: H-C-C-H."""

    def __init__(self, coupling_path=None, nmr_active_atoms=None, subset_atoms=None):
        super(J3HH, self).__init__(coupling_path=coupling_path,
                                   nmr_active_atoms=nmr_active_atoms,
                                   subset_atoms=subset_atoms)

    def couplings(self, carbon_atom):
        """Generate list of possible J3HH couplings.

        :param carbon_atom: Carbon atom.
        :type atom: :class:`ctfile.Atom`
        :return: List of couplings.
        :rtype: :py:class:`list`
        """
        couplings = []
        for neighbor_carbon_atom in carbon_atom.neighbor_carbon_atoms:
            if len(neighbor_carbon_atom.neighbor_hydrogen_atoms) > 0:
                coupling_path = [
                    [ResonanceAtom(atom=atom, isotope=self.nmr_active_atoms.get(atom.atom_symbol, '')) for atom in carbon_atom.neighbor_hydrogen_atoms],
                    [ResonanceAtom(atom=carbon_atom, isotope=self.nmr_active_atoms.get(carbon_atom.atom_symbol, ''))],
                    [ResonanceAtom(atom=neighbor_carbon_atom, isotope=self.nmr_active_atoms.get(neighbor_carbon_atom.atom_symbol, ''))],
                    [ResonanceAtom(atom=atom, isotope=self.nmr_active_atoms.get(atom.atom_symbol, '')) for atom in neighbor_carbon_atom.neighbor_hydrogen_atoms]
                ]
                couplings.append(self.__class__(coupling_path=coupling_path,
                                                nmr_active_atoms=self.nmr_active_atoms,
                                                subset_atoms=self.subset_atoms))
        return couplings

    def is_resonance_compatible(self, resonance):
        """Test if ``J3HH`` coupling is compatible with resonance.

        :param resonance: Subclass of :class:`~isoenum.nmr.Coupling`.
        :type resonance: :class:`~isoenum.nmr.Coupling`
        :return: True if compatible, False otherwise.
        :rtype: :py:obj:`True` or `False`
        """
        if set(resonance.hydrogen_coupling_path_repr[0]) == set(self.hydrogen_coupling_path_repr[0]):
            return True
        else:
            return False


class HResonance(Coupling):
    """Hydrogen resonance."""

    def __init__(self, coupling_path=None, nmr_active_atoms=None, subset_atoms=None):
        super(HResonance, self).__init__(coupling_path=coupling_path,
                                         nmr_active_atoms=nmr_active_atoms,
                                         subset_atoms=subset_atoms)

    def couplings(self, carbon_atom):
        """Generate resonance.

        :param carbon_atom: Carbon atom.
        :type atom: :class:`ctfile.Atom`
        :return: List of couplings.
        :rtype: :py:class:`list`
        """
        couplings = []
        if len(carbon_atom.neighbor_hydrogen_atoms) > 0:
            coupling_path = [
                [ResonanceAtom(atom=atom, isotope=self.nmr_active_atoms.get(atom.atom_symbol, '')) for atom in carbon_atom.neighbor_hydrogen_atoms],
                [ResonanceAtom(atom=carbon_atom, isotope=self.nmr_active_atoms.get(carbon_atom.atom_symbol, ''))]
            ]
            couplings.append(self.__class__(coupling_path=coupling_path,
                                            nmr_active_atoms=self.nmr_active_atoms,
                                            subset_atoms=self.subset_atoms))
        return couplings


class NMRExperiment(object):
    """NMR experiment."""

    def __init__(self, name, couplings, decoupled, default_coupling_definitions):
        """NMR experiment initializer.

        :param str name: NMR experiment name (type).
        :param list couplings: List of allowed coupling types.
        :param list decoupled: List of decoupled elements.
        :param list default_coupling_definitions: List of default coupling definitions.
        """
        self.name = name
        self.couplings = couplings
        self.decoupled = decoupled
        self.default_coupling_definitions = default_coupling_definitions

        if not couplings:
            self.coupling_definitions = [coupling for coupling
                                         in self.default_coupling_definitions]
        else:
            self.coupling_definitions = [self.default_coupling_definitions[coupling.upper()] for coupling
                                         in couplings if coupling in self.default_coupling_definitions]

        if decoupled:
            self.coupling_definitions = [coupling for coupling in self.coupling_definitions
                                         if not set(coupling.nmr_active_atoms).intersection([element.upper()
                                                                                             for element in decoupled])]

    def generate_coupling_combinations(self, molfile, subset=False):
        """Generate possible J couplings for a ``Molfile``."""
        raise NotImplementedError('Subclass must implement method.')


class NMR1D1H(NMRExperiment):
    """1D 1H NMR experiment."""

    def __init__(self, name, couplings=None, decoupled=None,
                 default_coupling_definitions=(HResonance(nmr_active_atoms={'H': '1'}, subset_atoms={'H': '2'}),
                                               J1CH(nmr_active_atoms={'C': '13', 'H': '1'}, subset_atoms={'H': '2'}),
                                               J2HH(nmr_active_atoms={'H': '1'}, subset_atoms={'H': '2'}),
                                               J3HH(nmr_active_atoms={'H': '1'}, subset_atoms={'H': '2'}))):
        """NMR experiment initializer.
        
        :param str name: NMR experiment name (type).
        :param list couplings: List of allowed coupling types.
        :param list decoupled: List of decoupled elements.
        """
        super(NMR1D1H, self).__init__(name=name, couplings=couplings, decoupled=decoupled,
                                      default_coupling_definitions=default_coupling_definitions)

        self.possible_coupling_type_combinations = [combination for combination in utils.all_combinations(self.coupling_definitions)
                                                    if any(isinstance(coupling, HResonance)
                                                    for coupling in combination)]

    def generate_coupling_combinations(self, molfile, subset=False):
        """Generate 1D 1H NMR experiment J couplings for a ``Molfile``.
        
        :param molfile: ``Molfile`` object.
        :type molfile: :class:`ctfile.Molfile`
        :param subset: Generate couplings subsets?
        :type subset: :py:obj:`True` or :py:obj:`False`
        :return: List of possible couplings for a given ``Molfile``.
        :rtype: :py:class:`list`
        """
        valid_coupling_combinations = []
        valid_subset_coupling_combinations = []

        for carbon_atom in molfile.carbon_atoms:
            if len(carbon_atom.neighbor_hydrogen_atoms) > 0:
                for coupling_type_combination in self.possible_coupling_type_combinations:
                    coupling_combination_lists = []
                    for coupling_type in coupling_type_combination:
                        couplings = coupling_type.couplings(carbon_atom)

                        if couplings:
                            coupling_combination_lists.append(utils.all_combinations(couplings))

                        for product in itertools.product(*coupling_combination_lists):
                            coupling_combination = list(more_itertools.flatten(product))

                            if coupling_combination not in valid_coupling_combinations:
                                valid_coupling_combinations.append(coupling_combination)

        if subset:
            for coupling_combination in valid_coupling_combinations:
                subsets = [coupling.subset for coupling in coupling_combination]

                for product in itertools.product(*subsets):
                    subset_product = list(product)
                    subset_product = [coupling for coupling in subset_product
                                      if coupling.is_resonance_compatible(subset_product[0])]

                    if subset_product not in valid_subset_coupling_combinations:
                        valid_subset_coupling_combinations.append(subset_product)

            valid_coupling_combinations = valid_subset_coupling_combinations
        return valid_coupling_combinations


class NMR1DCHSQC(NMR1D1H):
    """1D 13C HSQC NMR experiment."""

    def __init__(
        self,
        name,
        couplings=None,
        decoupled=None,
        default_coupling_definitions=(
            HResonance(nmr_active_atoms={'C': '13', 'H': '1'}, subset_atoms={'H': '2'}),
            J1CH(nmr_active_atoms={'C': '13', 'H': '1'}, subset_atoms={'H': '2'}),
            J2HH(nmr_active_atoms={'C': '13', 'H': '1'}, subset_atoms={'H': '2'}),
            J3HH(nmr_active_atoms={'H': '1'}, subset_atoms={'H': '2'}))):
        """

        :param name:
        :param couplings:
        :param decoupled:
        :param default_coupling_definitions:
        """
        super(NMR1DCHSQC, self).__init__(name=name, couplings=couplings, decoupled=decoupled,
                                         default_coupling_definitions=default_coupling_definitions)


def create_nmr_experiment(name, couplings=None, decoupled=None):
    """Create NMR experiment upon provided experiment type.
    
    :param str name: NMR experiment type.
    :param list couplings: List of coupling types.
    :param list decoupled: List of elements.
    :return: NMR experiment.
    :rtype: :class:`~isoenum.nmr.NMRExperiment`.
    """
    nmr_experiment_types = {
        '1D1H': NMR1D1H,
        '1DCHSQC': NMR1DCHSQC
    }

    try:
        return nmr_experiment_types[name.upper()](name=name, couplings=couplings, decoupled=decoupled)
    except KeyError:
        raise ValueError('Unknown nmr experiment type: "{}"'.format(name))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
isoenum.utils
~~~~~~~~~~~~~

This module provides reusable utility functions.
"""

import itertools

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse


def is_url(path):
    """Test if path represents a valid URL.
    
    :param str path: Path to file.
    :return: True if path is valid url string, False otherwise.
    :rtype: :py:obj:`True` or :py:obj:`False`
    """
    try:
        parse_result = urlparse(path)
        return all((parse_result.scheme, parse_result.netloc, parse_result.path))
    except ValueError:
        return False


def all_combinations(items):
    """Generate combinations of variable size.
    
    :param items: Sequence of items.
    :return: List of all combinations.
    :rtype: :py:class:`list`
    """
    combinations = []
    for rsize in range(1, len(items) + 1):
        combinations.extend(list(itertools.combinations(items, rsize)))
    return combinations



#!/usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
isoenum.iso_property
~~~~~~~~~~~~~~~~~~~~

This module contains a function to generate isotopic properties block from 
provided atom labeling schema.
"""

from ctfile.conf import ctab_properties_conf


def create_iso_property(labeling_schema, ctab_properties_conf=ctab_properties_conf, version='V2000'):
    """Create "ISO" property block.

    :param list labeling_schema: Labeling schema containing info about atom, isotope, and its position.
    :param str version: Version of ``CTfile``.
    """
    print(labeling_schema)


    iso_property_fmt = ctab_properties_conf[version]['ISO']['fmt']
    atom_iso_property = '{}  1{}{}'
    ctab_properties_block = [atom_iso_property.format(iso_property_fmt,
                                                      str(i['position']).rjust(4),
                                                      str(i['isotope']).rjust(4))
                             for i in labeling_schema]
    return ctab_properties_block


#
#
# import ctfile
# from pprint import pprint
#
# with open('tests/example_data/iso305.mol', 'r') as inf:
#     ctf = ctfile.load(inf)
#
# ctf.print_file()
#
#



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
isoenum.cli
~~~~~~~~~~~

Isotopic enumerator (isoenum) command-line interface

Usage:
    isoenum -h | --help
    isoenum --version
    isoenum name (<path-to-ctfile-file-or-inchi-file-or-inchi-string>) 
                 [--specific=<isotope:element:position>...] 
                 [--all=<isotope:element>...] 
                 [--enumerate=<isotope:element:min:max>...] 
                 [--complete | --partial] 
                 [--ignore-iso]
                 [--format=<format>]
                 [--output=<path>]
                 [--verbose]
    isoenum ionize (<path-to-ctfile-file-or-inchi-file-or-inchi-string>)
                   (--state=<element:position:charge>...)
                   [--format=<format>]
                   [--output=<path>]
    isoenum nmr (<path-to-ctfile-file-or-inchi-file-or-inchi-string>)
                [--type=<experiment-type>]
                [--jcoupling=<name>...]
                [--decoupled=<element>...]
                [--format=<format>]
                [--output=<path>]
                [--subset]
                [--verbose]
    isoenum vis (<path-to-ctfile-file-or-inchi-file-or-inchi-string>) 
                (--format=<format>)
                (--output=<path>)

Options:
    -h, --help                                 Show this screen.
    --verbose                                  Print more information.
    -v, --version                              Show version.
    -a, --all=<isotope:element>                Specify element and isotope, e.g. -a 13:C or --all=13:C
    -s, --specific=<isotope:element:position>  Specify element, isotope and specific position,
                                               e.g. -s 13:C:1 or --specific=13:C:1.
    -e, --enumerate=<isotope:element:min:max>  Enumerate all isotopically-resolved CTfile or InChI,
                                               e.g. -e 13:C:2:4 or --enumerate=13:C:2:4
    -c, --complete                             Use complete labeling schema, i.e. every atom must specify
                                               "ISO" property, partial labeling schema will be used otherwise
                                               for specified labeling information only.
    -p, --partial                              Use partial labeling schema, i.e. generate labeling schema
                                               from the provided labeling information.
    -i, --ignore-iso                           Ignore existing "ISO" specification in the CTfile or InChI.
    -f, --format=<format>                      Format of output: inchi, mol, sdf, csv, json [default: inchi].
    -o, --output=<path>                        Path to output file.
    -t, --type=<experiment-type>               Type of NMR experiment [default: 1D1H].
    -j, --jcoupling=<type>                     Allowed J couplings.
    -d, --decoupled=<element>                  Turn off J coupling for a given element.
    -z, --state=<element:position:charge>      Create ionized form of InChI from neutral molecule, 
                                               e.g. N:6:+1, O:8:-1.
    --subset                                   Create atom subsets for each resonance.
"""

from __future__ import print_function, division, unicode_literals

import csv
import io
import os
import sys

from . import api
from .conf import output_formats_conf


def cli(cmdargs):
    """Process command-line arguments.

    :param dict cmdargs: Command-line arguments.
    :return: None.
    :rtype: :py:obj:`None`
    """
    path_or_id = cmdargs['<path-to-ctfile-file-or-inchi-file-or-inchi-string>']

    if cmdargs['name']:
        sdfile = api.iso(path_or_id=path_or_id,
                         specific_opt=cmdargs['--specific'],
                         all_opt=cmdargs['--all'],
                         enumerate_opt=cmdargs['--enumerate'],
                         complete_opt=cmdargs['--complete'],
                         ignore_iso_opt=cmdargs['--ignore-iso'])

        create_output(sdfile=sdfile, path=cmdargs['--output'], file_format=cmdargs['--format'])

    elif cmdargs['ionize']:
        atom_states = cmdargs['--state']
        sdfile = api.chg(path_or_id=path_or_id, atom_states=atom_states)
        create_output(sdfile=sdfile, path=cmdargs['--output'], file_format=cmdargs['--format'])

    elif cmdargs['nmr']:
        experiment_type = cmdargs['--type']
        decoupled = [element.upper() for element in cmdargs['--decoupled']]
        couplings = [coupling.upper() for coupling in cmdargs['--jcoupling']]
        subset = cmdargs['--subset']

        sdfile = api.iso_nmr(path_or_id=path_or_id, experiment_type=experiment_type,
                             couplings=couplings, decoupled=decoupled, subset=subset)

        create_output(sdfile=sdfile, path=cmdargs['--output'], file_format=cmdargs['--format'])

    elif cmdargs['vis']:
        api.visualize(path_or_id=path_or_id,
                      output_path=cmdargs['--output'],
                      output_format=cmdargs['--format'])


def _unpack(param):
    """Unpack command-line option.

    :param list param: List of isotopes.
    :return: List of unpacked isotopes.
    :rtype: :py:class:`list`
    """
    options = []
    for option_str in param:
        options.extend(option_str.split(','))
    return options


def save_output(outputstr, path, file_format):
    """Save output results into file or print to stdout.

    :param str outputstr: Output string. 
    :param str path: Where to save results.
    :param str file_format: File format to create file extension.
    :return: None.
    :rtype: :py:obj:`None`
    """
    if path is not None:
        dirpath, basename = os.path.split(os.path.normpath(path))
        filename, extension = os.path.splitext(basename)

        if not extension or extension.lower() not in output_formats_conf:
            extension = '.{}'.format(file_format)

        filename = '{}{}'.format(filename, extension)
        filepath = os.path.join(dirpath, filename)

        if dirpath and not os.path.exists(dirpath):
            raise IOError('Directory does not exist: "{}"'.format(dirpath))

        with open(filepath, 'w') as outfile:
            print(outputstr, file=outfile)
    else:
        print(outputstr, file=sys.stdout)


def create_output(sdfile, path=None, file_format='inchi'):
    """Create output containing conversion results.

    :param sdfile: ``SDfile`` instance.
    :type sdfile: :class:`~ctfile.ctfile.SDfile`.
    :param str path: Path to where file will be saved. 
    :param str file_format: File format: 'inchi', 'mol', 'sdf', 'json', or 'csv'.
    :return: None.
    :rtype: :py:obj:`None`
    """
    file_format = file_format.lower()

    if file_format not in output_formats_conf:
        raise ValueError('Unknown output format: "{}".\n'
                         'Available formats are: {}'.format(file_format, output_formats_conf))

    if file_format in {'sdf', 'mol'}:
        save_output(outputstr=sdfile.writestr(file_format='ctfile'),
                    path=path,
                    file_format=file_format)

    elif file_format in {'json'}:
        save_output(outputstr=sdfile.writestr(file_format='json'),
                    path=path,
                    file_format=file_format)

    elif file_format in {'inchi'}:
        output = []
        for entry_id in sdfile:
            output.extend([inchi.strip() for inchi in sdfile[entry_id]['data']['InChI']])

        save_output(outputstr='\n'.join(output),
                    path=path,
                    file_format=file_format)

    elif file_format in {'csv'}:
        outputstr = io.StringIO()
        csvwriter = csv.writer(outputstr, delimiter='\t')

        for entry_id in sdfile:
            csv_data = []
            for data_id in sdfile[entry_id]['data']:
                value = ' + '.join([item.strip() for item in sdfile[entry_id]['data'][data_id]])
                csv_data.append(value)
            csvwriter.writerow(csv_data)

        save_output(outputstr=outputstr.getvalue(),
                    path=path,
                    file_format=file_format)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import docopt

from . import cli
from . import __version__


def main():
    args = docopt.docopt(cli.__doc__, version=__version__)
    cli.cli(args)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""This package provides routines to generate isotopically-resolved ``InChI`` 
(International Chemical Identifier) from non-isotopically labeled ``CTfile``
formatted files (e.g. ``Molfiles``, ``SDfiles``) or non-isotopically labeled 
InChI identifiers.

This package includes the following modules:

``cli``
    This module provides the command-line interface for the ``isoenum`` package.

``labeling``
    This module provides functions for generating a labeling schema.

``nmr``
    This module provides descriptions of coupling combinations that
    could be observed within NMR experiments.

``openbabel``
    This module provides functions to call the Open Babel software to convert 
    between ``InChI`` and ``CTfile`` formatted files. 

``conf``
    This module provides the processing of configuration files necessary for 
    isotopic enumerator.

``fileio``
    This module provides functions for generating ``CTfile`` objects and
    convert ``CTfile`` objects into ``InChI``.

``utils``
    This module provides reusable utility functions.
"""

import logging

from . import fileio
from . import api
from . import openbabel


__version__ = '0.4.0'


try:  # Python 2/3 compatibility code
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass


# Setting default logging handler
logging.getLogger(__name__).addHandler(NullHandler())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
import re
from setuptools import setup, find_packages


if sys.argv[-1] == "publish":
    os.system("python3 setup.py sdist")
    os.system("twine upload dist/*")
    sys.exit()


def readme():
    with open("README.rst") as readme_file:
        return readme_file.read()


def find_version():
    with open("isoenum/__init__.py", "r") as fd:
        version = re.search(
            r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
        ).group(1)
    if not version:
        raise RuntimeError("Cannot find version information")
    return version


REQUIRES = [
    "docopt >= 0.6.2",
    "ctfile >= 0.1.6",
    "more_itertools >= 4.1.0",
    "requests >= 2.18.4",
    "packaging >= 19.1",
]


setup(
    name="isoenum",
    version=find_version(),
    author="Andrey Smelter",
    author_email="andrey.smelter@gmail.com",
    description="InChI generator.",
    keywords="InChI InChIKey isoenum isotopic enumerator",
    license="BSD",
    url="https://github.com/MoseleyBioinformaticsLab/isoenum",
    packages=find_packages(),
    package_data={"isoenum": ["config_files/*.json"]},
    platforms="any",
    long_description=readme(),
    install_requires=REQUIRES,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    entry_points={"console_scripts": ["isoenum = isoenum.__main__:main"]},
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
isoenum.api
~~~~~~~~~~~

This module provides routines to augment ``CTfile`` objects 
with additional information and used by ``isoenum`` package CLI. 
"""

import tempfile
from collections import Counter
from collections import defaultdict
from collections import OrderedDict

import more_itertools

from . import fileio
from . import labeling
from . import nmr
from . import openbabel
from .conf import isotopes_conf


def iso(path_or_id, specific_opt=None, all_opt=None, enumerate_opt=None, complete_opt=False, ignore_iso_opt=False):
    """Create isotopically-resolved ``SDfile``.

    :param str path_or_id: Path to ``CTfile`` or file identifier.
    :param list specific_opt: List of isotopes per specific element type and position.
    :param list all_opt: List of isotopes for specific element type. 
    :param list enumerate_opt: List of isotopes to perform enumeration. 
    :param complete_opt: Identify if every element need to have isotope information.
    :type complete_opt: py:obj:`True` or py:obj:`False`  
    :param ignore_iso_opt: Ignore existing isotope information or not.
    :type ignore_iso_opt: py:obj:`True` or py:obj:`False`
    :return: instance of ``SDfile``.
    :rtype: :class:`ctfile.ctfile.SDfile`
    """
    ctfile = fileio.create_ctfile(path_or_id=path_or_id)
    sdfile = fileio.create_empty_sdfile_obj()

    if specific_opt is None:
        specific_opt = []

    if all_opt is None:
        all_opt = []

    if enumerate_opt is None:
        enumerate_opt = []

    for molfile in ctfile.molfiles:
        existing_opt = ['{}:{}:{}'.format(atom.isotope, atom.atom_symbol, atom.atom_number)
                        for atom in molfile.atoms if atom.isotope]

        specific_iso = _check_specific_opt(isotopes=specific_opt, isotopes_conf=isotopes_conf, ctfile=molfile)
        existing_iso = _check_specific_opt(isotopes=existing_opt, isotopes_conf=isotopes_conf, ctfile=molfile)
        all_iso = _check_all_opt(isotopes=all_opt, isotopes_conf=isotopes_conf, ctfile=molfile)
        enumerate_iso = _check_enumerate_opt(enumerate_opt=enumerate_opt, all_iso=all_iso,
                                             isotopes_conf=isotopes_conf, ctfile=molfile)

        labeling_schemas = labeling.create_labeling_schema(complete_labeling_schema=complete_opt,
                                                           ignore_existing_isotopes=ignore_iso_opt,
                                                           all_iso=all_iso, specific_iso=specific_iso,
                                                           existing_iso=existing_iso, enumerate_iso=enumerate_iso,
                                                           isotopes_conf=isotopes_conf, ctfile=molfile)

        for labeling_schema in labeling_schemas:
            sdfile_data = OrderedDict()
            new_molfile = fileio.create_ctfile_from_ctfile_str(ctfile_str=molfile.writestr(file_format='ctfile'))

            for isotope_spec in labeling_schema:
                new_molfile.atom_by_number(atom_number=isotope_spec['atom_number']).isotope = isotope_spec['isotope']

            sdfile_data.setdefault('InChI', []).append('{}'.format(fileio.create_inchi_from_ctfile_obj(new_molfile)))
            sdfile.add_molfile(molfile=new_molfile, data=sdfile_data)

    return sdfile


def chg(path_or_id, atom_states):
    """Create ``SDfile`` with charge information.

    :param str path_or_id: Path to ``CTfile`` or file identifier. 
    :param list atom_states: List of charges for specific elements. 
    :return: instance of ``SDfile``.
    :rtype: :class:`ctfile.ctfile.SDfile`
    """
    ctfile = fileio.create_ctfile(path_or_id=path_or_id)
    sdfile = fileio.create_empty_sdfile_obj()

    for molfile in ctfile.molfiles:
        sdfile_data = OrderedDict()
        for state in atom_states:
            try:
                atom_symbol, atom_number, charge = state.split(':')
            except ValueError:
                raise ValueError('Incorrect ionization specification, use "element:position:charge" format.')

            molfile.add_charge(atom_symbol=atom_symbol, atom_number=atom_number, charge=charge)

        new_molfile = fileio.create_ctfile_from_ctfile_str(ctfile_str=molfile.writestr(file_format='ctfile'))
        sdfile_data.setdefault('InChI', []).append('{}'.format(fileio.create_inchi_from_ctfile_obj(ctf=molfile)))
        sdfile.add_molfile(molfile=new_molfile, data=sdfile_data)

    return sdfile


def iso_nmr(path_or_id, experiment_type, couplings, decoupled, subset):
    """Create isotopically-resolved ``SDfile`` assuming specific NMR experiment type.

    :param str path_or_id: Path to ``CTfile`` or file identifier.
    :param str experiment_type: NMR experiment type (1D1H of 1DCHSQC). 
    :param list couplings: What couplings to include?
    :param list decoupled: What elements are decoupled?
    :param subset: Create subsets?
    :type subset: py:obj:`True` or py:obj:`False`
    :return: instance of ``SDfile``.
    :rtype: :class:`ctfile.ctfile.SDfile`
    """
    ctfile = fileio.create_ctfile(path_or_id=path_or_id)
    sdfile = fileio.create_empty_sdfile_obj()
    nmr_experiment = nmr.create_nmr_experiment(name=experiment_type, couplings=couplings, decoupled=decoupled)

    for molfile in ctfile.molfiles:
        molfile = fileio.normalize_ctfile_obj(molfile)
        coupling_combinations = nmr_experiment.generate_coupling_combinations(molfile=molfile, subset=subset)

        for coupling_combination in coupling_combinations:
            sdfile_data = OrderedDict()

            ctab_iso_layer = []
            for coupling in coupling_combination:
                for atom in more_itertools.flatten(coupling.coupling_path):
                    if atom.atom_symbol in coupling.nmr_active_atoms:
                        ctab_iso_layer.append({'atom_number': str(atom.atom_number),
                                               'isotope': str(atom.isotope)})

                sdfile_data.setdefault('CouplingType', []).append(coupling.name)

            new_molfile = create_new_molfile(molfile=molfile, ctab_iso_layer=ctab_iso_layer)

            sdfile_data.setdefault('InChI', []).append(
                '{}'.format(fileio.create_inchi_from_ctfile_obj(ctf=new_molfile)))
            sdfile.add_molfile(molfile=new_molfile, data=sdfile_data)

    annotate_me_groups(ctfile=sdfile)
    return sdfile


def coupling_descr(coupling_types):
    """Concatenate a list of coupling types into coupling description string.

    :param coupling_types: List of coupling types.
    :return: Coupling description string.
    :rtype: :py:class:`str`
    """
    return " + ".join([coupling_type.split("]")[1] for coupling_type in coupling_types])


def create_new_molfile(molfile, ctab_iso_layer):
    """Create new `Molfile` instance with new isotopic labeling specification.

    :param ctfile: `Molfile` instance.
    :type ctfile: :class:`~ctfile.ctfile.Molfile`
    :param list ctab_iso_layer: Isotopic layer specification.
    :return: New `Molfile` instance.
    :rtype: :class:`~ctfile.ctfile.Molfile`
    """
    new_molfile = fileio.create_ctfile_from_ctfile_str(ctfile_str=molfile.writestr(file_format='ctfile'))
    for isotope_spec in ctab_iso_layer:
        new_molfile.atom_by_number(atom_number=isotope_spec['atom_number']).isotope = isotope_spec['isotope']
    return new_molfile


def create_aggregate_molfile(ctfile, entry_ids):
    """Create aggregate `Molfile` that has isotopic labeling from all associated entry ids.

    :param ctfile: `SDfile` instance.
    :type ctfile: :class:`~ctfile.ctfile.SDfile`
    :param list entry_ids: List of entry ids within `SDfile` instance.
    :return: Tuple of new `Molfile` and data associated with it.
    :rtype: :py:class:`tuple`
    """
    ctab_iso_layer = []
    coupling_types = set()
    sdfile_data = OrderedDict()

    for entry_id in entry_ids:
        molfile = ctfile[entry_id]['molfile']

        for coupling in ctfile[entry_id]['data']['CouplingType']:
            coupling_types.add(coupling)

        for atom in molfile.atoms:
            iso_property = {'atom_number': str(atom.atom_number),
                            'isotope': str(atom.isotope)}
            if iso_property not in ctab_iso_layer:
                ctab_iso_layer.append(iso_property)
            else:
                continue

    new_molfile = create_new_molfile(molfile=molfile, ctab_iso_layer=ctab_iso_layer)
    sdfile_data['CouplingType'] = sorted(coupling_types, key=lambda x: (x.split("]")[1], x))
    sdfile_data['InChI'] = [fileio.create_inchi_from_ctfile_obj(ctf=new_molfile)]

    return new_molfile, sdfile_data


def create_inchi_groups(ctfile):
    """Organize `InChI` into groups based on their identical `InChI` string and similar coupling type.

    :param ctfile: `SDfile` instance.
    :type ctfile: :class:`~ctfile.ctfile.SDfile`
    :return: Dictionary of related `InChI` groups.
    :rtype: :rtype: :py:class:`dict`
    """
    inchi_groups = defaultdict(list)
    for entry_id, entry in ctfile.items():
        inchi_str = entry["data"]["InChI"][0]
        descr = coupling_descr(coupling_types=entry["data"]["CouplingType"])
        inchi_groups[(inchi_str, descr)].append(entry_id)
    return inchi_groups


def annotate_me_groups(ctfile):
    """Annotate magnetically equivalent (have the same `InChI`) coupling type groups.

    :param ctfile: `SDfile` instance.
    :type ctfile: :class:`~ctfile.ctfile.SDfile`
    :return: None.
    :rtype: :py:obj:`None`
    """
    inchi_groups = create_inchi_groups(ctfile)
    me_ids = {inchi_str: group_id for inchi_str, group_id in zip(inchi_groups.keys(), list(range(1, len(inchi_groups) + 1)))}

    for data in ctfile.sdfdata:
        inchi_str = data["InChI"][0]
        descr = coupling_descr(coupling_types=data["CouplingType"])
        me_group_id = "ME{}".format(me_ids[(inchi_str, descr)])
        data["MEGroup"] = [me_group_id]

    for inchi_group, entry_ids in inchi_groups.items():
        if len(entry_ids) > 1:
            new_molfile, sdfile_data = create_aggregate_molfile(ctfile=ctfile, entry_ids=entry_ids)
            me_group_id = "MEA{}".format(me_ids[inchi_group])
            sdfile_data["MEGroup"] = [me_group_id]
            ctfile.add_molfile(molfile=new_molfile, data=sdfile_data)
        else:
            continue


def _check_specific_opt(isotopes, isotopes_conf, ctfile):
    """Check if `specific` option is consistent.

    :param list isotopes: Option that specifies isotopes.
    :param dict isotopes_conf: Default isotopes configuration.
    :param ctfile: Instance of ``Molfile``. 
    :type ctfile: :class:`~ctfile.ctfile.Molfile`
    :return: :py:class:`dict` with position specific labeling if consistent, raises error otherwise.
    :rtype: :py:class:`dict` or :py:class:`ValueError`
    """
    position_atom = defaultdict(list)
    allowed_atom_symbols = [atom.atom_symbol for atom in ctfile.atoms]
    positions = [atom.atom_number for atom in ctfile.atoms]
    ctfile_position_atom = dict(zip(positions, allowed_atom_symbols))
    specific_iso = {}

    for isotope_str in isotopes:
        try:
            isotope, atom, position = isotope_str.split(':')
        except ValueError:
            raise ValueError('Incorrect isotope specification, use "isotope:element:position" format.')

        if atom not in allowed_atom_symbols:
            raise ValueError('Incorrect atom "{}" provided.'.format(atom))

        if isotope not in isotopes_conf[atom]["isotopes"]:
            raise ValueError('Incorrect isotope "{}" provided for atom "{}".'.format(isotope, atom))

        if ctfile_position_atom[position] != atom:
            raise ValueError('There is no "{}" atom at position "{}"'.format(atom, position))

        position_atom[position].append('{}-{}'.format(atom, isotope))
        specific_iso[position] = {'atom_symbol': atom, 'isotope': isotope, 'atom_number': position}

    if all(len(isotopes) == 1 for isotopes in position_atom.values()):
        return specific_iso
    else:
        raise ValueError('"specific" option can only specify single isotope per atom position.')


def _check_all_opt(isotopes, isotopes_conf, ctfile):
    """Check if `all` option is consistent.

    :param list isotopes: Option that specifies isotopes.
    :param dict isotopes_conf: Default isotopes configuration.
    :param ctfile: Instance of ``Molfile``. 
    :type ctfile: :class:`~ctfile.ctfile.Molfile`
    :return: :py:class:`dict` with position specific labeling if consistent, raises error otherwise.
    :rtype: :py:class:`dict` or :py:class:`ValueError`
    """
    allowed_atom_symbols = [atom.atom_symbol for atom in ctfile.atoms]
    positions = [atom.atom_number for atom in ctfile.atoms]
    position_atom = dict(zip(positions, allowed_atom_symbols))
    atom_isotope = defaultdict(list)
    all_iso = {}

    for isotope_str in isotopes:
        try:
            isotope, atom = isotope_str.split(':')
        except ValueError:
            raise ValueError('Incorrect isotope specification, use "isotope:element" format.')

        if atom not in allowed_atom_symbols:
            raise ValueError('Incorrect atom "{}" provided.'.format(atom))

        if isotope not in isotopes_conf[atom]["isotopes"]:
            raise ValueError('Incorrect isotope "{}" provided for atom "{}".'.format(isotope, atom))

        atom_isotope[atom].append(isotope)

        for position, ctfile_atom in position_atom.items():
            if atom == ctfile_atom:
                all_iso[position] = {'atom_symbol': atom, 'isotope': isotope, 'atom_number': position}

    if all(len(isotopes) == 1 for isotopes in atom_isotope.values()):
        return all_iso
    else:
        raise ValueError('"all" option can only specify single isotope per atom type.')


def _check_enumerate_opt(enumerate_opt, all_iso, isotopes_conf, ctfile):
    """Check if `enumerate` option is consistent.

    :param list enumerate_param: Option that specifies isotopes.
    :param dict all_param: Check for consistency, `enumerate` are not compatible with `all` for the same atom.
    :param dict isotopes_conf: Default isotopes configuration. 
    :param ctfile: Instance of ``Molfile``. 
    :type ctfile: :class:`~ctfile.ctfile.Molfile`
    :return: :py:class:`list` with position specific labeling if consistent, raises error otherwise.
    :rtype: :py:class:`list` or :py:class:`ValueError`
    """
    atom_counter = Counter([atom.atom_symbol for atom in ctfile.atoms])
    allowed_atom_symbols = [atom.atom_symbol for atom in ctfile.atoms]
    all_param_atoms = [entry['atom_symbol'] for entry in all_iso.values()]
    enumerate_iso = []

    for isotope_str in enumerate_opt:
        try:
            isotope, atom, min_count, max_count = isotope_str.split(':')

        except ValueError:
            try:
                isotope, atom, max_count = isotope_str.split(':')
                min_count = 0

            except ValueError:
                try:
                    isotope, atom = isotope_str.split(':')
                    max_count = atom_counter[atom]
                    min_count = 0

                except ValueError:
                    raise ValueError('Incorrect isotope specification, use "isotope:element:min:max",'
                                     '"isotope:element:max" or "isotope:element" format.')

        enumerate_iso.append({'atom_symbol': atom, 'isotope': isotope,
                              'min': int(min_count), 'max': int(max_count)})

        if atom in all_param_atoms:
            raise ValueError('"enumerate" and "all" options are not compatible for atom: "{}"'.format(atom))

        if atom not in allowed_atom_symbols:
            raise ValueError('Incorrect atom "{}" provided.'.format(atom))

        if isotope not in isotopes_conf[atom]['isotopes']:
            raise ValueError('Incorrect isotope "{}" provided for atom "{}".'.format(isotope, atom))

        if int(max_count) > atom_counter[atom]:
            raise ValueError('Incorrect count "{}" provided for atom "{}".'.format(max_count, atom))

    return enumerate_iso


def visualize(path_or_id, output_path, output_format='svg', **options):
    """Visualize ``CTfile`` object.

    :param str path_or_id: Path to ``CTfile`` or file identifier.
    :type ctfile: :class:`~ctfile.ctfile.Molfile` or :class:`~ctfile.ctfile.SDfile`
    :param str output_format: Image output format. 
    :param str output_path: Image output path.
    """
    ctfile = fileio.create_ctfile(path_or_id=path_or_id)

    if output_format not in {'svg', 'png'}:
        output_format = 'svg'

    with tempfile.NamedTemporaryFile(mode='w') as tempfh:
        tempfh.write(ctfile.writestr(file_format='ctfile'))
        tempfh.flush()

        openbabel.convert(input_file_path=tempfh.name,
                          output_file_path=output_path,
                          input_format='mol',
                          output_format=output_format,
                          **options)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
isoenum.fileio
~~~~~~~~~~~~~~

This module provides routines to generate ``CTfile`` objects and
convert ``CTfile`` objects into ``InChI`` and vice versa.
"""

import os
import tempfile
import logging

import ctfile
import requests

from . import openbabel
from . import utils


logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())


def create_ctfile(path_or_id, xyx_coordinates='--gen2D', explicit_hydrogens='-h'):
    """Guess what type of path is provided, i.e. is it existing file in ``Molfile`` format, 
    existing file in ``SDfile`` file format, existing file containing ``InChI`` string, 
    or ``InChI`` string and tries to create ``CTfile`` object.

    :param str path_or_id: Path to ``Molfile``, ``SDfile``, ``InChI``, or ``InChI`` string.
    :param str xyx_coordinates: Option that generates x, y, z coordinates (e.g., "--gen2D" or "--gen3D").
    :param str explicit_hydrogens: Option that makes hydrogens atoms explicit when generating ``CTfile`` object.
    :return: Subclass of :class:`~ctfile.ctfile.CTfile` object.
    :rtype: :class:`~ctfile.ctfile.Molfile` or :class:`~ctfile.ctfile.SDfile`
    """
    if os.path.isfile(path_or_id):
        with open(path_or_id, 'r') as infile:
            try:
                return ctfile.load(infile)
            except IndexError:
                return create_ctfile_from_identifier_file(path_or_id,
                                                          xyz_coordinates=xyx_coordinates,
                                                          explicit_hydrogens=explicit_hydrogens)

    elif utils.is_url(path_or_id):
        try:
            return ctfile.read_file(path_or_id)
        except IndexError:
            identifier_str = requests.get(path_or_id).text
            ctf = create_ctfile_from_identifier_str(identifier_str=identifier_str,
                                                    xyx_coordinates=xyx_coordinates,
                                                    explicit_hydrogens=explicit_hydrogens)
    else:
        ctf = create_ctfile_from_identifier_str(identifier_str=path_or_id,
                                                xyx_coordinates=xyx_coordinates,
                                                explicit_hydrogens=explicit_hydrogens)
    if ctf:
        return ctf
    else:
        raise ValueError('Cannot create "CTfile" object, empty object.')


def create_ctfile_from_ctfile_str(ctfile_str):
    """Create ``CTfile`` object from ``CTfile`` string.

    :param str ctfile_str: ``CTfile`` string. 
    :return: Subclass of :class:`~ctfile.ctfile.CTfile` object.
    :rtype: :class:`~ctfile.ctfile.CTfile` 
    """
    return ctfile.loadstr(ctfile_str)


def create_ctfile_from_identifier_file(path, output_format='mol', **options):
    """Create ``CTfile`` instance from ``InChI`` or ``SMILES`` identifier file.

    :param str path: Path to file containing ``InChI`` or ``SMILES`` identifier.
    :param str output_format: Output file format used by Open Babel to generate ``CTfile``.
    :param options: Additional options to be passed to Open Babel.
    :return: Subclass of :class:`~ctfile.ctfile.CTfile` object.
    :rtype: :class:`~ctfile.ctfile.CTfile`
    """
    with open(path, 'r') as infile:
        input_format = guess_identifier_format(identifier_str=infile.read())

    with tempfile.NamedTemporaryFile() as tempfh:
        openbabel.convert(input_file_path=path,
                          output_file_path=tempfh.name,
                          input_format=input_format,
                          output_format=output_format,
                          **options)

        with open(tempfh.name, 'r') as infile:
            return ctfile.load(infile)


def create_ctfile_from_identifier_str(identifier_str, output_format='mol', **options):
    """Create ``CTfile`` instance from ``InChI`` or ``SMILES`` identifier string.

    :param str identifier_str: ``InChI`` or ``SMILES`` identifier string.
    :param str output_format: Output file format used by Open Babel to generate ``CTfile``.
    :param options: Additional options to be passed to Open Babel.
    :return: Subclass of :class:`~ctfile.ctfile.CTfile` object.
    :rtype: :class:`~ctfile.ctfile.CTfile`
    """
    with tempfile.NamedTemporaryFile(mode='w') as tempfh:
        tempfh.write(identifier_str)
        tempfh.flush()
        return create_ctfile_from_identifier_file(path=tempfh.name, output_format=output_format, **options)


def guess_identifier_format(identifier_str):
    """Guess identifier format.

    :param str identifier_str: Chemical identifier string.
    :return: 'inchi' or 'smiles' string.
    :rtype: :py:class:`str`
    """
    if identifier_str.startswith('InChI='):
        return 'inchi'
    else:
        return 'smiles'


def create_inchi_from_ctfile_obj(ctf, **options):
    """Create ``InChI`` from ``CTfile`` instance.

    :param ctf: Instance of :class:`~ctfile.ctfile.CTfile`.
    :type ctf: :class:`~ctfile.ctfile.CTfile`
    :return: ``InChI`` string.
    :rtype: :py:class:`str`
    """
    # apply fixed hydrogen layer when atom charges are present
    atom_charges = [atom.charge for atom in ctf.atoms if atom.charge != '0']
    if atom_charges:
        options.update({'fixedH': '-xF'})

    with tempfile.NamedTemporaryFile(mode='w') as moltempfh, tempfile.NamedTemporaryFile(mode='r') as inchitempfh:
        moltempfh.write(ctf.writestr(file_format='ctfile'))
        moltempfh.flush()
        openbabel.convert(input_file_path=moltempfh.name,
                          output_file_path=inchitempfh.name,
                          input_format='mol',
                          output_format='inchi',
                          **options)
        inchi_result = inchitempfh.read()
    return inchi_result.strip()


def normalize_ctfile_obj(ctf, xyx_coordinates='--gen2D', explicit_hydrogens='-h'):
    """Normalize ``CTfile`` object.

    :param ctf: ``CTfile`` object.
    :type ctf: :class:`~ctfile.ctfile.CTfile`
    :return: Normalized ``CTfile`` object.
    :rtype: :class:`~ctfile.ctfile.CTfile`
    """
    identifier_str = create_inchi_from_ctfile_obj(ctf)
    return create_ctfile_from_identifier_str(identifier_str=identifier_str,
                                             xyx_coordinates=xyx_coordinates,
                                             explicit_hydrogens=explicit_hydrogens)


def create_empty_sdfile_obj():
    """Create empty ``SDfile`` object.

    :return: ``SDfile`` object.
    :rtype: :class:`~ctfile.ctfile.SDfile`
    """
    return ctfile.SDfile()


def create_empty_molfile_obj():
    """Create empty ``Molfile`` object.

    :return: ``Molfile`` object.
    :rtype: :class:`~ctfile.ctfile.Molfile`
    """
    return ctfile.Molfile()


def create_svg_str(inchi_str, **options):
    """Create SVG string from ``InChI`` identifier string.

    :param str inchi_str: ``InChI`` identifier.
    :param options: Additional options to be passed to Open Babel.
    :return: SVG XML code.
    :rtype: :py:class:`str`
    """
    with tempfile.NamedTemporaryFile(mode='w') as inchi_tempfh, tempfile.NamedTemporaryFile(mode='r') as svg_tempfh:
        inchi_tempfh.write(inchi_str)
        inchi_tempfh.flush()

        openbabel.convert(input_file_path=inchi_tempfh.name,
                          output_file_path=svg_tempfh.name,
                          input_format='inchi',
                          output_format='svg',
                          **options)
        svg_str = svg_tempfh.read()
    return svg_str


def circular_consistency_test(inchi_str):
    """Perform conversion from ``InChI`` string to ``Molfile`` and back to
    ``InChI`` string.

    :param str inchi_str: ``InChI`` string.
    :return: None
    :rtype: :py:obj:`None`
    """
    original_inchi_str = inchi_str.strip()

    with tempfile.NamedTemporaryFile(mode='w') as temp_inchi_fh, \
         tempfile.NamedTemporaryFile(mode='w') as temp_mol_fh, \
         tempfile.NamedTemporaryFile(mode='w') as temp_converted_inchi_fh:

        temp_inchi_fh.write(original_inchi_str)
        temp_inchi_fh.flush()

        options = {}
        if inchi_str.startswith('InChI=1/'):
            options['fixedh'] = '-xF'

        openbabel.convert(input_file_path=temp_inchi_fh.name, output_file_path=temp_mol_fh.name,
                          input_format='inchi', output_format='mol', gen3D='--gen3D')

        openbabel.convert(input_file_path=temp_mol_fh.name, output_file_path=temp_converted_inchi_fh.name,
                          input_format='mol', output_format='inchi', **options)

        with open(temp_converted_inchi_fh.name, 'r') as inf:
            converted_inchi_str = inf.read().strip()

    if original_inchi_str != converted_inchi_str:
        logger.warning('WARNING: Circular conversion test (InChI to molfile to InChI) did not pass.'
                       'Original InChI: {}'
                       'Converted InChI: {}'.format(original_inchi_str, converted_inchi_str))

